(** License: GPLv3. *)

(** A concrete element for a multithreaded program is a set of states, where each state is a tuple having a global state and an array of local states. *)

module type SharedLocalType = sig
  type shared
  type component (** Type of local states *)
  val compare_shared: shared->shared->int 
    (** [compare_shared x y] returns signum(x-y) *)
  val compare_component: component->component->int
    (** [compare_component x y] returns signum(x-y) *)
  val string_of_shared: shared->string
    (** prints a representation of a shared state into a string *)
  val string_of_component: component->string
    (** prints a representation of a component state into a string *)
end


module type S = sig
  type shared
    (** The type of shared state part. *)

  type component
    (** The type of each component. *)
  
  type setOfComponent
    (** The type of a set of components. *)

  type setOfShared
    (** The type of a set of shared states. *)

  type state
    (** the type of a conrecte element *)
    
  type setOfStates
    (** the type of a set of concrete elements *)
    
  val empty_setOfStates: setOfStates
    (** An empty set of states *)

  val is_empty_setOfStates: setOfStates->bool
    (** Checks whether a given set of states is empty *)

  val single_state: shared -> component array -> state
    (** creates a single state from the shared and the local part *)

  val singleton_setOfStates: shared -> component array -> setOfStates
    (** creates a singleton from the shared and the local part *)
    
  val choose_state_from_setOfStates: setOfStates -> state
    (** Chooses any state from a set of states or raises Not_found if the 
	set is empty. *)

  val state_dimension: state->int
    (** Returns the number of locals of a state. *)

  val add_state_setOfStates: state->setOfStates->setOfStates
    (** Adds an element to the set of states *)
    
  val union_setOfStates_setOfStates: setOfStates->setOfStates->setOfStates
    (** Returns the union of two sets of states *)

  val intersection_setOfStates_setOfStates:
    setOfStates->setOfStates->setOfStates
    (** Returns the intersection of two sets of states *)

  val diff_setOfStates_setOfStates:
    setOfStates->setOfStates->setOfStates
    (** Returns the difference between two sets of states. *)

  val string_of_state:state->string
    (** Returns a string representation of a string. *)

  val string_of_setOfStates: setOfStates->string
    (** Returns a string representation of a set of states. *)

  val string_of_trace: setOfStates list -> string
    (** create a string representation of a list of states *)

  type transition
    (** Type of a single transition. *)
    
  type transitionSet
    (** Type of a set of transitions. *)

  val empty_transitionSet : transitionSet
    (** The empty transiton set*)

  val is_empty_transitionSet:transitionSet->bool
    (** Whether a set of transitions is empty *)

  val single_transition:shared->component->int->shared->component->transition
    (** Creating a single transition. *)

  val singleton_transition: 
    shared->component->int->shared->component->transitionSet
    (** Creates a singleton transition. *)

  val add_transition_transitionSet:transition->transitionSet->transitionSet
    (** Adds a transition to a set of transitions. *)

  val union_transitionSet_transitionSet:transitionSet->transitionSet->transitionSet
    (** Takes a union of two sets of transitions *)

  val apply_transitionSet: setOfStates -> transitionSet -> setOfStates
    (** [apply_transitionSet sos ts] returns a set of states which results 
	from applying all transitions from [ts] to [sos] and uniting the 
	results. *)

  val apply_reversed_transitionSet: setOfStates -> transitionSet -> setOfStates
    (** [apply_reversed_transitionSet sos ts] returns a set of states which 
	results from applying the reverse of all transitions from [ts] to [sos]
	and uniting the results. 
	The reverse of (g,l,i,g',l') is (g',l',i,g,l). *)

  val reverse_transitionSet: transitionSet -> transitionSet
    (** [reverse_transitionSet ts] returns the set
	of all transitions (g',l',i,g,l) such that (g,l,i,g',l') is in [ts]
    *)

  val string_of_transitionSet: transitionSet->string
    (** Convert a transition set into a string. *)

  type abstractElement
    (** The type of an element of the abstract domain. *)

  val least_abstractElement: abstractElement
    (** The least abstract element. *)

  val string_of_abstractElement:abstractElement->string
    (** Creating a string descrition of an abstract element. *)

  val abstract_setOfStates: setOfStates -> abstractElement
    (** Compute the abstraction of a set of states. *)

  val abstract_setOfStates_Minus_setOfStates: setOfStates->setOfStates->(abstractElement*setOfStates)
    (** Compute the abstraction of a difference between two sets of states. 
	Return the difference in the first component. Return a more optimized 
	version of the second set of states in the second component. *)

  val join_abstract_element_cartabs_setOfStates_minus_setOfStates: abstractElement->setOfStates->setOfStates->(abstractElement*setOfStates)
    (** [join_abstract_element_abstract_setOfStates_minus_setOfStates a a' e] 
	computes the join of a with (cartesianAbstraction(a'\e)) in the first 
	component. A more saturated version of e is returned in the second 
	component. *)

  val le_abstract_elements: abstractElement -> abstractElement -> bool
  (** Tests whether an abstract element is less than or equal to another 
      abstract element. *)

  val join_abstract_elements:abstractElement->abstractElement->abstractElement
    (** Creates a join of two abstract elements. *)

  val concretize_abstractElement_union_setOfStates: abstractElement->setOfStates->setOfStates
    (** Concretizes an abstract element and unites with a given set of tuples.
    *)

  exception ExceptionSetNotFound
    (** Gets raised if a precondition for generating an exception set is 
	not satisfied. *)

  val exception_set: abstractElement-> setOfStates -> setOfStates ->setOfStates
  (** [exception_set c b f] where c is an abstract element and b and f are
      sets of states finds some subset e of b such that the intersection of f 
      with the cartesian abstraction of (c union (b\e)) is empty. It assumes 
      that the intersection of f with (c union b) is empty. *)

  val no_mutex_setOfStates: setOfComponent->setOfComponent->setOfShared->int->setOfStates
  (** [no_mutex_setOfStates inside all globs n] creates the set of states
      { (g,x1,...,xn) | g in globs and exists i unequal j: xi in inside 
      and xj in inside and for all k unequal i,j we have xk in all}. *)

  val no_assertion_setOfStates: setOfComponent->setOfComponent->setOfShared->int->setOfStates
  (** [no_assertion_setOfStates inside all globs n] creates the set of states
      { (g,x1,...,xn) | g in globs and exists i : xi in inside 
      and for all k unequal i we have xk in all}. *)
end


module MakeOrderedLocal(SL:SharedLocalType): (CartesianProduct.OrderedTypeWithPrinting with type t=SL.component)

module MakeOrderedShared(SL:SharedLocalType): (Set.OrderedType with type t=SL.shared)

module Make (SL:SharedLocalType) : (S with type shared=SL.shared and type component=SL.component and type setOfComponent=Set.Make(MakeOrderedLocal(SL)).t and type setOfShared=Set.Make(MakeOrderedShared(SL)).t)

