(** License: GPLv3. *)

(** Bijective maps over two ordered types.

   This module implements bijective maps, given a total ordering functions
   over the domain and a total ordering function over the range.
   All operations over maps are purely applicative (no side-effects).
   The implementation uses two maps, for which currently (2008) searching
   and insertion take time logarithmic in the size of the map.
*)

module type PairOfOrderedTypes = sig
  type left_type (** Type of the elements in the domain of the map. *)
  type right_type (** Type of the elements in the range of the map. *)
  val compare_left: left_type -> left_type ->int
  (** A total ordering function over the elements of the domain. 
      This is a two-argument function [f] such that
      [f e1 e2] is zero if the [e1] and [e2] are equal,
      [f e1 e2] is strictly negative if [e1] is smaller than [e2],
      and [f e1 e2] is strictly positive if [e1] is greater than [e2].
      Example: a suitable ordering function is the generic structural
      comparison function {!Pervasives.compare}.  *)
  val compare_right:  right_type -> right_type ->int
  (** A total ordering function over the elements of the domain. 
      This is a two-argument function [f] such that
      [f e1 e2] is zero if the [e1] and [e2] are equal,
      [f e1 e2] is strictly negative if [e1] is smaller than [e2],
      and [f e1 e2] is strictly positive if [e1] is greater than [e2].
      Example: a suitable ordering function is the generic structural
      comparison function {!Pervasives.compare}.  *)
end
(** Input signature of the functor {!BijectiveMap.Make} *)

module type S = sig
  type left_key (** Type of the elements in the domain of the map. *)
  type right_key (** Type of the elements in the range of the map. *)
  type t (** type of bijective maps from left_key to right_key. *)
  val empty: t (** The empty bijection. *)
  val is_empty: t -> bool (** Test whether a bijection is empty or not.*)
  val add: left_key -> right_key -> t -> t (** [add x y m] returns a bijection containing the same bindings as [m], plus a binding of [x] to [y]. If x was already bound in bound in m, its previous bindings disappears, if [y] was already bound in [m], its previous binding disappears. *)
  val find_left: left_key -> t -> right_key (** [find_left x m] returns m(x), if [x] is in the domain of [m], otherwise it raises [Not_found]. *)
  val find_right: right_key -> t -> left_key (** [find_right y m] returns the preimage of [y] under [m], or raises [Not_found] if no such binding exists. *)
  val mem_left: left_key -> t -> bool (** [mem_left x m] returns true if there is pair in [m] with [x] in the first component *)
  val mem_right: right_key -> t -> bool (** [mem_right y m] returns true if there is pair in [m] with [y] in the second component *)
  val max_left: t->left_key (** Return the greatest element of the domain (with respect to the [P.compare_left] ordering), or raise Not_found if the bijection is empty. *)
  val max_right: t->right_key (** Return the greatest element of the range (with respect to the [P.compare_right] ordering), or raise Not_found if the bijection is empty. *)
end
(** Output signature of the functor {!BijectiveMap.Make} *)

module Make (P : PairOfOrderedTypes) : S with type left_key = P.left_type and type right_key = P.right_type
(** Functor building an implementation of the bijective map structure given two totally ordered types. *)
