(* Written by Alexander Malkis, 2009. Licence: GPLv3 *)

(** A Cartesian product of a user-given fixed dimension over a single user-given ordered type. Implementation: array of sets. For each underlying type, the empty Cartesian product has a unique representation, which has zero empty components. *)

module type OrderedTypeWithPrinting = sig
  type t
  val compare: t->t->int
  val to_string: t->string
end

module type C = sig
  type componentType
  type setOfComponentType
  type t
  val empty: t
  val is_empty: t->bool
  val make: int -> setOfComponentType -> t
  val join: t-> t-> t
  val inter: t-> t-> t
  val compare: t-> t-> int
  val subset: t-> t-> bool
  val equal: t-> t-> bool
  val string_of_CP: t-> string

  type setOfTuples
  val empty_setOfTuples: setOfTuples
  val is_empty_setOfTuples: setOfTuples ->bool
  val singleton_setOfTuples: componentType array -> setOfTuples
  val choose_tuple_from_setOfTuples: setOfTuples -> componentType array
  val union_CP_setOfTuples: t-> setOfTuples-> setOfTuples
  val add_tuple_to_setOfTuples: componentType array ->setOfTuples->setOfTuples
  val union_setOfTuples_setOfTuples: setOfTuples->setOfTuples->setOfTuples
  val inter_setOfTuples_setOfTuples: setOfTuples->setOfTuples->setOfTuples
  val subset_setOfTuples_setOfTuples: setOfTuples->setOfTuples->bool
  val diff_setOfTuples_setOfTuples: setOfTuples->setOfTuples->setOfTuples
  val cartabs_diff_setOfTuples_setOfTuples: setOfTuples->setOfTuples->(t*setOfTuples)
  val join_CP_cartabs_diff_setOfTuples_setOfTuples: t->setOfTuples->setOfTuples->(t*setOfTuples)
  val cartabs_setOfTuples:setOfTuples->t
  val setOfTuples_from_CP: t->setOfTuples
  val apply_transition_setOfTuples: setOfTuples -> int -> componentType -> componentType -> setOfTuples
  exception ExceptionSetNotFound
  val exception_set_for_setOfTuples: t->setOfTuples->setOfTuples->setOfTuples
  val eager_exception_set_for_setOfTuples: t->setOfTuples->setOfTuples->setOfTuples
  val mutex_setOfTuples: setOfComponentType->setOfComponentType->int->setOfTuples
  val no_mutex_setOfTuples: setOfComponentType->setOfComponentType->int->setOfTuples
  val string_of_setOfTuples:setOfTuples->string
  val string_of_setOfTuples_possibly_with_brackets:setOfTuples->string
end

module Make(Ord:OrderedTypeWithPrinting) = struct
  type componentType=Ord.t

  module SetOfComponentType = Set.Make(Ord)

  type setOfComponentType=SetOfComponentType.t

  type t=setOfComponentType array

  (* There are no other representations of the empty cartesian product. *)
  let empty= Array.make 0 (SetOfComponentType.empty)

  (* Checks whether a product has an empty component and is thus empty itself. 
     Only for internal use. *)
  let is_any_component_of_CP_empty c =
    let arrlen=Array.length c
    in if arrlen<=0 then true else let j=ref(arrlen-1) and found = ref false in
      while (!j>=0 && not !found) do
	found:=(SetOfComponentType.is_empty c.(!j)) ; decr j; done;
      !found
      
  (* Checks whether a product is the empty product. *)
  let is_empty c = ((Array.length c)<=0)

  let canonize c = if is_empty c then empty else c

  (* Choose any tuple from a Cartesian product. Throws Not_found if the 
     product is empty. *)
  let rec choose_tuple_from_CP (cp:t) : componentType array=
    let len=Array.length cp in
      if len<=0 then raise Not_found 
      else let init=Array.make 1 (SetOfComponentType.choose cp.(0)) in
	if len=1 then init
	else let remainder_cp=Array.sub cp 1 (len-1) in 
	let remainder_chosen= choose_tuple_from_CP remainder_cp in
	  Array.append init remainder_chosen

  (* Returns the number of dimensions of a product. *)
  let get_dimension_of_CP cp =
    if is_empty cp then 0 else Array.length cp

  (* creates a string representing a set of component types *)
  let string_of_setOfComponentType s =
    "{"^(
      if SetOfComponentType.is_empty s then "" else
	let first=SetOfComponentType.min_elt s in 
	let rest=SetOfComponentType.remove first s in
	  SetOfComponentType.fold (fun c sofar -> sofar^","^(Ord.to_string c)) rest (Ord.to_string first)
    )^"}"

  (* creates a string description of a cartesian product *)
  let string_of_CP cp =
    let len=Array.length cp in if len<=0 then "{}" 
      else let last=cp.(len-1) in
      let rest=Array.sub cp 0 (len-1) in
	Array.fold_right (fun ae sofar -> (string_of_setOfComponentType ae)^"x"^sofar) rest (string_of_setOfComponentType last)

  (* create a set of tuples containing only one tuple *)
  let singleton_CP (t:componentType array):t =
    let len=Array.length t in
    let arr=Array.make len (SetOfComponentType.empty)
    in for i=0 to len-1 do arr.(i)<-(SetOfComponentType.singleton t.(i)) done;
      arr

  (* Creates a product from an array of sets. *)
  let make i s = 
    if SetOfComponentType.is_empty s then empty else Array.make i s

  (* Computes a componentwise union of two products. *)
  let join s1 s2 =
    let s1len=Array.length s1 in if s1len<=0 then s2 else
	let s2len=Array.length s2 in if s2len<=0 then s1 else
	    if s1len<>s2len then raise (Invalid_argument ("CartesianProduct.join: trying to take componentwise union of nonempty Cartesian products with dimensions "^(string_of_int s1len)^" and "^(string_of_int s2len)^".")) else
	      let r=Array.make s1len SetOfComponentType.empty in
		for i=0 to s1len - 1 do 
		  r.(i)<-SetOfComponentType.union s1.(i) s2.(i) done;
		r

  (* Computes a componentwise intersection of two products. *)
  let inter s1 s2 =
    let s1len=Array.length s1 in if s1len<=0 then empty else 
	let s2len=Array.length s2 in if (s2len<=0)||(s1len<>s2len) then empty
	  else 
	    let r=Array.make s1len SetOfComponentType.empty
	    and j=ref(s1len-1) and found_empty=ref false in
	      while ((!j>=0) && (not !found_empty)) do
		r.(!j)<-(SetOfComponentType.inter s1.(!j) s2.(!j));
		found_empty:=SetOfComponentType.is_empty r.(!j);
		decr j
	      done;
	      if !found_empty then empty else r

  (* Induces some total ordering on Cartesian products. *)
  let compare s1 s2 =
    let s1len=Array.length s1 and s2len=Array.length s2 in
      if s1len<s2len then -1 else if s1len>s2len then 1 else
	let i=ref (s1len-1) and found=ref 0 in 
	  while ((!i>=0) && (0 = !found)) do
	    found:=SetOfComponentType.compare s1.(!i) s2.(!i); decr i done;
	  !found

  (* Is one product a subset of another one. *)
  let subset s1 s2 =
(*    Debug.print_debug ("CartesianProduct.subset started on "^(string_of_CP s1)^" and "^(string_of_CP s2)^".\n"); *)
    let result=
      let s1len=Array.length s1 in if s1len<=0 then true else
	  let s2len=Array.length s2 in if s1len<>s2len then  false else 
	      let i=ref(s1len-1) and still_subset=ref true in
		while (!i>=0) && !still_subset do
		  still_subset:=SetOfComponentType.subset s1.(!i) s2.(!i); 
		  decr i
		done;
		!still_subset
    in (* Debug.print_debug ("CartesianProduct.subset finished with "^(string_of_bool result)^".\n"); *)
      result

  (* Checks whether two products are equal. *)
  let equal (s1:t) (s2:t) =
    let s1len=Array.length s1 and s2len=Array.length s2 in
      if s1len<=0 then (if s2len<=0 then true else false)
      else if s1len<>s2len then false else
	let i=ref(s1len-1) and still_equal=ref true in 
	  while (!i>=0) && !still_equal do
	    still_equal:=SetOfComponentType.equal s1.(!i) s2.(!i); 
	    decr i  done;
	  !still_equal

  (* Creates a Cartesian product outside^i x inside x outside^{n-i-1} *)
  let mutex_CP (inside:Set.Make(Ord).t) (outside:Set.Make(Ord).t) (i:int) (n:int) : t =
    if (SetOfComponentType.is_empty inside) || 
      ((SetOfComponentType.is_empty outside)&&(n>=2)) then empty else
	let r=Array.make n outside in ((r.(i)<-inside);r)

  (* [no_mutex_CP inside all i j n] creates the Cartesian product 
     all^k x inside x all^{l-k-1} x inside x all^{n-l-1}
     where k=min{i,j},l=max{i,j}, i unequal j*)
  let no_mutex_CP inside all i j n = 
    if i=j then raise (Invalid_argument "CartesianProduct.no_mutex_CP expects different component numbers.\n")
    else if (SetOfComponentType.is_empty inside) || 
      ((SetOfComponentType.is_empty all)&&(n>=3)) then empty else
	let r=Array.make n all in ((r.(i)<-inside);r.(j)<-inside;r)
				    

  (* A list of Cartesian products, representing its union. Empty products are 
     not allowed in this list. *)
  type setOfTuples = t list

  (* The unique representation of the empty set of tuples. *)
  let empty_setOfTuples:setOfTuples =[]

  (* Check whether a set of tuples is empty or not *)
  let is_empty_setOfTuples (s:setOfTuples)= match s with []->true | _ ->false

  (* Choose any tuple from a set of tuples. *)
  let choose_tuple_from_setOfTuples = function
      [] -> raise Not_found
    | h::t -> choose_tuple_from_CP h

  (* creates a representation of a set of tuples *)
  let rec string_of_setOfTuples = function
      [] -> "{}"
    | [h] -> string_of_CP h
    | h::t -> (string_of_CP h)^"u"^(string_of_setOfTuples t)
    
  (* creates a representation of a set of tuples, possibly with brackets *)
  let string_of_setOfTuples_possibly_with_brackets sot = 
    let s= string_of_setOfTuples sot in
      if (List.length sot)>1 then "("^s^")" else s

  (* How to compare sets *)
  type noOfIncomparableComponents=One|MoreThanOne
  type subsetOrder=Less|Greater|Equal|Incomparable of noOfIncomparableComponents

  (* Convert a subset order into a string *)
  let string_of_subsetOrder = function
      Less -> "Less"
    | Greater -> "Greater"
    | Equal -> "Equal"
    | Incomparable One -> "Incomparable One"
    | Incomparable MoreThanOne -> "Incomparable MoreThanOne"

  (* compare two Cartesian products as sets *)
  let compare_cartprod_wrt_subset (c1:t) (c2:t) :subsetOrder =
    (* Debug.print_debug ("CartesianProduct.compare_cartprod_wrt_subset started on products "^(string_of_CP c1)^" and "^(string_of_CP c2)^".\n"); *)
    let result=
      (let len1=Array.length c1 and len2=Array.length c2
       in if len1<=0 then (if len2<=0 then Equal else Less)
	 else if len2<=0 then Greater
	 else if len1<>len2 then Incomparable MoreThanOne
	 else let i=ref(len1-1) and cmp_res_till_now=ref Equal in 
	   while (!i>=0) && (match !cmp_res_till_now with 
				 Incomparable MoreThanOne -> false
			       | _->true) 
	   do
	     (let one_less_two=SetOfComponentType.subset c1.(!i) c2.(!i) and
		  two_less_one=SetOfComponentType.subset c2.(!i) c1.(!i)
	      in let now= 
		  if one_less_two then 
		    (if two_less_one then Equal else Less) 
		  else
		    (if two_less_one then Greater else Incomparable One)
	      in (cmp_res_till_now:=
		    match (!cmp_res_till_now,now) with
			(Less,Less) -> Less
		      | (Less, Greater) -> Incomparable MoreThanOne
		      | (Greater,Less) -> Incomparable MoreThanOne
		      | (Greater,Greater)->Greater
		      | (Equal,x) -> x
		      | (x,Equal) -> x
		      | ((Incomparable _),_) -> Incomparable MoreThanOne
		      | (_,(Incomparable _)) -> Incomparable MoreThanOne
		 ));
	     decr i
	   done;
	   !cmp_res_till_now)
    in (*
	 Debug.print_debug ("CartesianProduct.compare_cartprod_wrt_subset finished with "^(string_of_subsetOrder result)^".\n"); *)
      result



  (* removes from s all those products which are subsets of s *)
  let rec remove_subsumed_by_CP (c:t) (s:setOfTuples) :setOfTuples =
    match s with
	[] -> []
      | h::t -> if subset h c then remove_subsumed_by_CP c t 
	else h::(remove_subsumed_by_CP c t)

  (* union_CP_setOfTuples_flagged c s returns the union of c and s in the 
     first component. The smaller Cartesian products are thrown out of s, 
     if a larger one is found in s, s is returned unchanged. If c differs from 
     some Cartesian product of s in only one component, both are merged.
     The second component of the result is true iff the represenation of the 
     union is different from the representation of s. *)
  let rec union_CP_setOfTuples_flagged (c:t) (s:setOfTuples) :setOfTuples*bool=
    (* Debug.print_debug ("CartesianProduct.union_CP_setOfTuples_flagged started on product "^(string_of_CP c)^" and set "^(string_of_setOfTuples s)^".\n"); *)
    let result=
      match s with
	  [] -> (if is_empty c then ([],false) else ([c],true))
	| h::t -> 
	    (match compare_cartprod_wrt_subset c h with
		 Less -> ((h::t),false)
	       | Greater -> ((c::(remove_subsumed_by_CP c t)),true)
	       | Equal -> ((h::t),false)
	       | Incomparable One ->
		   let c'=join c h in ((c'::(remove_subsumed_by_CP c' t)),true)
	       | Incomparable MoreThanOne ->
		 let (s',b)=union_CP_setOfTuples_flagged c t in ((h::s'),b)
	    )
    in
     (* Debug.print_debug ("CartesianProduct.union_CP_setOfTuples_flagged finished with result ("^(string_of_setOfTuples (fst result))^","^(string_of_bool (snd result))^").\n"); *)
      result

(*
  if subset c h then (h::t,false)
  else if subset h c then ((c::(remove_subsumed_by_CP c t)),true)
  else let (s',b)=union_CP_setOfTuples_flagged c t in ((h::s'),b)
*)

  (* create a set of tuples representing the inion of a cartesian product and 
     another set of tuples. *)
  let union_CP_setOfTuples cp sot = 
(*    Debug.print_debug ("CartesianProduct.union_CP_setOfTuples started on product "^(string_of_CP cp)^" and set "^(string_of_setOfTuples sot)^".\n"); *)
    let result= fst (union_CP_setOfTuples_flagged cp sot)
    in 
(*      Debug.print_debug ("CartesianProduct.union_CP_setOfTuples finished with result "^(string_of_setOfTuples result)^".\n");*)
      result

  (* Create a set of tuples consisting of only one tuple. *)
  let singleton_setOfTuples t = [singleton_CP t]
    
  (* add_tuple_to_setOfTuples t s returns s union {t} *)
  let add_tuple_to_setOfTuples (t:componentType array) (s:setOfTuples) =
    union_CP_setOfTuples (singleton_CP t) s

  (* union_setOfTuples_setOfTuples_flagged s1 s2 returns the union of s1 and 
     s2 in the first component. If s2 is an antichain, the first component of 
     the result is guaranteed to be an antichain. The second component of the 
     result is  true iff the represenation of the union is different from the 
     representation of s2. *)
  let rec union_setOfTuples_setOfTuples_flagged (s1:setOfTuples) (s2:setOfTuples) :setOfTuples*bool =
    match s1 with
	[] -> (s2,false)
      | h::t -> let (s2',b)=union_CP_setOfTuples_flagged h s2 
	in let (s2'',b')=union_setOfTuples_setOfTuples_flagged t s2'
	in (s2'',(b||b'))

  (* union_setOfTuples_setOfTuples s1 s2 returns the union of s1 and 
     s2. If s2 is an antichain, the result is guaranteed to be an antichain. *)
  let union_setOfTuples_setOfTuples s1 s2 = 
    fst (union_setOfTuples_setOfTuples_flagged s1 s2)

  (* Given X=X1x...xXn, Y=Y1x...xYn, 1<=i<=n, 
     resolve_on_given_component X Y i computes 
     (X1 inter Y1)x...x(X(i-1) inter Y(i-1))x
     x(Xi union Yi)x
     x(X(i+1) inter Y(i+1))x...x(Xn inter Yn)
  *)
  let resolve_on_given_component (x:t) (y:t) (i:int) :t =
    let len1=Array.length x and len2=Array.length y 
    in if len1<=0 then 
	(if (len2<=0)||(len2>=2)||(len2=1 && i<>0) then empty else y)
      else if len2<=0 then
	(if (len1>=2)||(len1=1 && i<>0) then empty else x)
      else
	if(len1<>len2) then raise (Invalid_argument ("The function resolve_on_given_component was given arrays of different lengths "^(string_of_int len1)^" and "^(string_of_int len2)^".\n")) else if i<0 then raise (Invalid_argument ("The function resolve_on_given_component was called on a negative component index to be resolved on: "^(string_of_int i)^".\n")) else if i>=len1 then raise (Invalid_argument ("The function resolve_on_given_component was called on component number "^(string_of_int i)^", but the largest component index is"^(string_of_int (len1-1))^".\n")) else 
	  let r=Array.make len1 SetOfComponentType.empty 
	  and j=ref 0 
	  and empty_found=ref false
	  in 
	    (while (!j<i) && (not !empty_found) do
	       (let intr=SetOfComponentType.inter x.(!j) y.(!j) in
		  (empty_found:=SetOfComponentType.is_empty intr;
		   if not !empty_found then ((r.(!j)<-intr);incr j) else ()));
	     done);
	    if not !empty_found then 
	      ((r.(i)<-(SetOfComponentType.union x.(i) y.(i)));incr j) else(); 
	    (while (!j<len1) && not !empty_found do
	       (let intr=SetOfComponentType.inter x.(!j) y.(!j) in
		  (empty_found:=SetOfComponentType.is_empty intr;
		   if not !empty_found then ((r.(!j)<-intr);incr j) else ()));
	     done); 
	    if !empty_found then empty else r

  (* resolve_on_all_components x y computes the list representing the union
     of Cartesian products due to resolution on all components. *)
  let resolve_on_all_components (x:t) (y:t) : setOfTuples =
    let len1=Array.length x and len2=Array.length y 
    in if len1<=0 then 
	(if (len2<=0)||(len2>=2) then empty_setOfTuples else [y])
      else if len2<=0 then
	(if (len1>=2) then empty_setOfTuples else [x])
      else
	if(len1<>len2) then raise (Invalid_argument ("The function resolve_on_given_component was given arrays of different lengths "^(string_of_int len1)^" and "^(string_of_int len2)^".\n"))	
	else let rec resolve_on_all_components_ge i accu=
	  if i<0 then accu else
	    let cp=resolve_on_given_component x y i
	    in resolve_on_all_components_ge (i-1) (union_CP_setOfTuples cp accu)
	in resolve_on_all_components_ge (len1-1) []
	       
  (* resolves a given product with each product in a list. 
     Returns the list of resolvents. *)
  let rec resolve_CP_with_setOfCP c s :setOfTuples =
    match s with
	[] -> []
      | h::t ->
	  let sofar=resolve_CP_with_setOfCP c t
	  in let res_c_h=resolve_on_all_components c h 
	  in union_setOfTuples_setOfTuples res_c_h sofar

  (* resolves a set of products with a set of other products, 
     by resolving each pair*)
  let rec resolve_setOfCP_setOfCP s1 s2 : setOfTuples=
    match s1 with
	[] -> []
      | h::t -> union_setOfTuples_setOfTuples
	  (resolve_CP_with_setOfCP h s2)
          (resolve_setOfCP_setOfCP t s2)
	  
  (* resolves a set once with itself. Returns the resolvents in the first 
     components. The second component is true iff the set is changed. *)
  let resolve_setOfCP_once s : setOfTuples*bool =
    (* Debug.print_debug ("CartesianProduct.resolve_setOfCP_once started on "^(string_of_setOfTuples s)^".\n"); *)
    let result=
      union_setOfTuples_setOfTuples_flagged (resolve_setOfCP_setOfCP s s) s
    in (* Debug.print_debug ("CartesianProduct.resolve_setOfCP_once finished with ("^(string_of_setOfTuples (fst result))^","^(string_of_bool (snd result))^").\n"); *) result
  
  (* remove_subsumed s makes an antichain from a set of Cartesian products *)
  let rec remove_subsumed (s:setOfTuples) :setOfTuples =
    match s with
	[] -> []
      | h::tl -> union_CP_setOfTuples h (remove_subsumed tl)

  (* saturate_setOfCP s returns the list containing all maximal products inside s and an undefinite number of products from s. *)
  let rec saturate_setOfCP s =
    (* Debug.print_debug ("CartesianProduct.saturate_setOfCP started on "^(string_of_setOfTuples s)^".\n"); *)
    let result=
      let (r,b)= resolve_setOfCP_once s 
      in if b then saturate_setOfCP r else r
    in (* Debug.print_debug "CartesianProduct.saturate_setOfCP finished.\n"; *)
result


  (* saturates the set of products with subsumtion. Returns the list of all
     maximal products inside s *)
  let saturate_setOfCP_with_subsumtion s: setOfTuples =
    saturate_setOfCP (remove_subsumed s)


  (* inter_CP_setOfTuples c s computes the intersection of c and s, 
     represented as an antichain of Cartesian products *)
  let inter_CP_setOfTuples (c:t) (s:setOfTuples) : setOfTuples = 
    List.fold_left  
      (fun sofar next_cp -> union_CP_setOfTuples (inter c next_cp) sofar) [] s 

  (* inter_setOfTuples_setOfTuples s1 s2 computes the intersection of 
     s1 and s2, represented as a union of Cartesian products. *)
  let inter_setOfTuples_setOfTuples s1 s2 =
(* Debug.print_debug "CartesianProduct.inter_setOfTuples_setOfTuples started.\n"; *)
    let result = List.fold_left (fun sofar_union curr_cp -> union_setOfTuples_setOfTuples (inter_CP_setOfTuples curr_cp s2) sofar_union) [] s1
    in 
(* Debug.print_debug "CartesianProduct.inter_setOfTuples_setOfTuples finished.\n"; *)
      result



  (* subset_CP_setOfTuples_flagged c s checks whether a product is contained in a set of tuples, represented as a union of products. During checking, saturation occurs, so a more saturated representation of the same union is in the second component of the returned pair. The first component contains the answer to the subset question. *)
  let subset_CP_setOfTuples_flagged (c:t) (s:setOfTuples) : bool*setOfTuples = 
    let restricted_saturated_s=saturate_setOfCP_with_subsumtion (inter_CP_setOfTuples c s)
    in ((List.exists (fun next_cp -> subset c next_cp) restricted_saturated_s),
	restricted_saturated_s)

(* restrict_CP c i elem returns the cartesian product c' such that if in c' {elem} is inserted at the position i, then the resulting cartesian product is a subset of c. *)
  let restrict_CP  (c:t) (i:int) (r:componentType) =
    let len = Array.length c in
      if len<=0 then empty else
	if i<0 then raise (Invalid_argument ("CartesianProduct.restrict_CP was required to do restiction on a negative component "^(string_of_int i)))
	else if i>=len then raise (Invalid_argument ("CartesianProduct.restrict_CP was required to do restiction on a too large component "^(string_of_int i)^" while the last components is "^(string_of_int (len-1))^".\n"))
	else if SetOfComponentType.mem r c.(i) then
	  Array.append (Array.sub c 0 i) (Array.sub c (i+1) (len-i-1))
	else empty

  (* diff_CP_CP cp1 cp2 returns the difference of two products. 
     It is assumed that both have the same dimension. *)
  let diff_CP_CP (cp1:t) (cp2:t) :setOfTuples =
    let rec add_dimension i sofar =
      if i<0 then sofar else
	let diff_i = SetOfComponentType.diff cp1.(i) cp2.(i)
	in if SetOfComponentType.is_empty diff_i then 
	    add_dimension (i-1) sofar
	  else
	    let new_arr=Array.copy cp1 in
	    let new_sofar=((new_arr.(i) <- diff_i);
			   union_CP_setOfTuples new_arr sofar)
	    in add_dimension (i-1) new_sofar
    in add_dimension ((Array.length cp1)-1) empty_setOfTuples
      
  (* diff_setOfTuples_CP sot cp returns the difference between a set of tuples
     sot and a product cp *)
  let diff_setOfTuples_CP (sot:setOfTuples) (cp:t) :setOfTuples =
(*
        let _= Debug.print_debug ("diff_setOfTuples_CP started on "^(string_of_setOfTuples sot)^" and "^(string_of_CP cp)^".\n") in let result=
*)
    let rec diff_setOfTuples_aux sofar = function
	[] -> sofar
      | (cp1::tail) -> 
	  let new_s=union_setOfTuples_setOfTuples (diff_CP_CP cp1 cp) sofar in
	    diff_setOfTuples_aux new_s tail
    in diff_setOfTuples_aux empty_setOfTuples sot
(*
        in let _= Debug.print_debug ("diff_setOfTuples_CP finished with "^(string_of_setOfTuples result)^".\n") in result
*)

  (* diff_setOfTuples_setOfTuples sot1 sot2 returns the difference between 
     two sets of tuples sot1\sot2 *)
  let rec diff_setOfTuples_setOfTuples sot = function
      [] -> sot
    | cp::tail -> diff_setOfTuples_setOfTuples (diff_setOfTuples_CP sot cp) tail

(* restrict_setOfTuples s i elem returns the set of all those tuples t such that if in t the element elem is inserted at position i, then the resulting tuple is in s. *)
  let restrict_setOfTuples s i elem =
    List.fold_left (fun sofar next_c -> union_CP_setOfTuples (restrict_CP next_c i elem) sofar) [] s

  (* Compute Product_j \in {1,..,n}\{i} A_j *)
  let product_without_component (cae:t) (i:int) =
    Array.append (Array.sub cae 0 i) 
      (Array.sub cae (i+1) ((Array.length cae)-i-1))

  (* unrestrict_CP c i elem returns 
     product_{j<i} c_j x {elem} x product_{j>i} c_j*)
  let unrestrict_CP c i elem =
    let len=Array.length c in
    let res=Array.make (len+1) SetOfComponentType.empty in
      for j=0 to (i-1) do res.(j)<-c.(j) done ;
      res.(i)<-SetOfComponentType.singleton elem;
      for j=i+1 to len do res.(j)<-c.(j-1) done ;
      res

  (* unrestrict_setOfTuples s i elem returns a set of tuples t who all have 
     elem at the ith component so that the remainder of the tuple 
     (i.e. t\{(i,elem)}) is in s *)
  let unrestrict_setOfTuples s i elem =
    List.fold_left (fun sofar next_c -> union_CP_setOfTuples (unrestrict_CP next_c i elem) sofar) [] s


  (* cartabs_diff_CP_minus_setOfTuples c s returns a pair. Its first component is the smallest product containing the difference c\s. Its second component is a more saturated verion of s. *)
  let cartabs_diff_CP_minus_setOfTuples (c:t) (s:setOfTuples) :(t*setOfTuples)=
    let len=Array.length c in 
      if len<=0 then (empty,s) else
	let s=saturate_setOfCP s in
	let res=Array.make len SetOfComponentType.empty 
	in let i=ref(len-1) and empty_found=ref false in 
	  while ((!i>=0) && (not !empty_found)) do
	    let restricted_c=product_without_component c !i
	    and restricted_s_i=restrict_setOfTuples s !i 
	    in let (res_i,s)=
		(SetOfComponentType.fold 
		   (fun elt (sofar_comp,sofar_new_s) ->
		      let restricted_s_i_elt = restricted_s_i elt 
		      in let (b,new_restricted_s_i_elt)=
			  subset_CP_setOfTuples_flagged restricted_c restricted_s_i_elt
		      in let unrestricted_new_s_delta=
			  unrestrict_setOfTuples new_restricted_s_i_elt !i elt
		      in ((if b then sofar_comp 
			   else (SetOfComponentType.add elt sofar_comp)),
			  union_setOfTuples_setOfTuples unrestricted_new_s_delta sofar_new_s
			 )
		   ) c.(!i) (SetOfComponentType.empty,s))
	    in if SetOfComponentType.is_empty res_i then empty_found:=true 
	      else ((res.(!i)<- res_i);decr i)
	  done;
	  ((if !empty_found then empty else res),s)
	    
	    
  (* join_CP_cartabs_diff_CP_minus_setOfTuples c cp s returns a pair. Its first
     component is the join of c with the smallest product containing the 
     difference c\s. Its second component is a more saturated verion of s. *)
  let join_CP_cartabs_diff_CP_minus_setOfTuples (c:t) (cp:t) (s:setOfTuples) :(t*setOfTuples)=
    let cplen=Array.length cp in if cplen<=0 then (c,s)
      else let len=Array.length c in if len<=0 then 
	cartabs_diff_CP_minus_setOfTuples cp s 
      else if len<>cplen then raise (Invalid_argument ("CartesianProduct.join_CP_cartabs_diff_CP_minus_setOfTuples was given products of different positive lengths "^(string_of_int len)^" and "^(string_of_int cplen)^".\n"))
      else let s=saturate_setOfCP s in
      let res=Array.make len SetOfComponentType.empty 
      in 
	(for i=0 to len-1 do
	   (let restricted_cp=product_without_component cp i
	    and restricted_s_i=restrict_setOfTuples s i 
	    and ci=c.(i)
	    in let (res_i,s)=
		(SetOfComponentType.fold 
		   (fun elt (sofar_comp,sofar_new_s) ->
		      let restricted_s_i_elt = restricted_s_i elt 
		      in let (b,new_restricted_s_i_elt)=
			  subset_CP_setOfTuples_flagged restricted_cp restricted_s_i_elt
		      in let unrestricted_new_s_delta=
			  unrestrict_setOfTuples new_restricted_s_i_elt i elt
		      in ((if b then sofar_comp 
			   else (SetOfComponentType.add elt sofar_comp)),
			  (union_setOfTuples_setOfTuples unrestricted_new_s_delta sofar_new_s)
			 )
		   ) (SetOfComponentType.diff cp.(i) ci) (ci,s))
	    in res.(i)<- res_i)
	 done;
	 (res,s))
	  
  (* cartabs_diff_setOfTuples_setOfTuples sot1 sot2 returns a cartesian product
     that is an abstraction of sot1\sot2 in the first component, 
     and a more saturated version of  sot2 in the second component. *)
  let cartabs_diff_setOfTuples_setOfTuples sot1 sot2 =
    List.fold_left 
      (fun (sofar_cp,sofar_sot2) cp -> 
	 let (cp',sofar_sot2')=cartabs_diff_CP_minus_setOfTuples cp sofar_sot2
	 in ((join sofar_cp cp'),sofar_sot2')) (empty,sot2) sot1

  (* cartabs_diff_setOfTuples_setOfTuples c sot1 sot2 returns the join of c
     with an abstraction of sot1\sot2 in the first component, 
     and a more saturated version of sot2 in the second component. *)
  let join_CP_cartabs_diff_setOfTuples_setOfTuples c sot1 sot2 =
(*    Debug.print_debug ("CartesianProduct.join_CP_cartabs_diff_setOfTuples_setOfTuples started on c="^(string_of_CP c)^" and sot1="^(string_of_setOfTuples sot1)^" and sot2="^(string_of_setOfTuples sot2)^".\n"); *)
    let result=
    let jcc=join_CP_cartabs_diff_CP_minus_setOfTuples c in
      List.fold_left
	(fun (sofar_cp,sofar_sot2) cp -> 
	   let (cp',sofar_sot2')= jcc cp sofar_sot2 in
	     ((join sofar_cp cp'),sofar_sot2')) (c,sot2) sot1
    in
(*    Debug.print_debug ("CartesianProduct.join_CP_cartabs_diff_setOfTuples_setOfTuples finished with result ("^(string_of_CP (fst result))^","^(string_of_setOfTuples (snd result))^").\n"); *)
      result

  (* cartesian abstraction of a set of tuples. *)
  let cartabs_setOfTuples sot =
    (* Debug.print_debug "CartesianProduct.cartabs_setOfTuples started.\n"; *)
    let result= fst(cartabs_diff_setOfTuples_setOfTuples sot empty_setOfTuples)
    in 
    (* Debug.print_debug "CartesianProduct.cartabs_setOfTuples finished.\n"; *)
      result

  (* subset_setOfTuples_setOfTuples_flagged s1 s2 returns a pair. The first 
     component is true iff s1 is a subset of s2. The second component 
     contains a more saturated version of s2.*)
  let rec subset_setOfTuples_setOfTuples_flagged s1 s2 = match s1 with
      [] -> (true,s2)
    | h::t -> let (b,s2new)=subset_CP_setOfTuples_flagged h s2 in
	if b then subset_setOfTuples_setOfTuples_flagged t s2new 
	else (false,s2new)

  (* subset_setOfTuples_setOfTuples s1 s2 returns true if s1 is a subset of s2,
     otherwise it returns false. *)
  let subset_setOfTuples_setOfTuples s1 s2 = 
    fst (subset_setOfTuples_setOfTuples_flagged s1 s2)

  (* Creates a set of tuples from a cartesian product. *)
  let setOfTuples_from_CP cp =
    if is_any_component_of_CP_empty cp then empty_setOfTuples else [cp]

  (* "apply_transition_CP c comp was is" returns a cartesian product which is 
     product_{j<comp} c_j x {is} x product_{j>comp} c_j
     if "was" is an element of c_comp, and it returns the empty product if 
     "was" is not an element of c_comp *)
  let apply_transition_CP (c:t) (comp:int) (was:componentType) (is:componentType) :t = 
    (* Debug.print_debug ("CartesianProduct.apply_transition_CP started with component number "^(string_of_int comp)^" and product "^(string_of_CP c)^".\n"); *)
    let result=
      if (comp>=Array.length c)||not(SetOfComponentType.mem was c.(comp)) then
	empty
      else
	  let res=Array.copy c
	  in (res.(comp)<-(SetOfComponentType.singleton is)); res
    in (* Debug.print_debug "CartesianProduct.apply_transition_CP finished.\n"; *)
      result

  (* "apply_transition_setOfTuples s comp was is" applies the transition 
     "was->is" to the component comp of each tuple in s and returnes the 
     obtained set, represented as a union over a list of products. *)
  let apply_transition_setOfTuples (s:setOfTuples) (comp:int) (was:componentType) (is:componentType) = 
      (* Debug.print_debug ("CartesianProduct.apply_transition_setOfTuples started on setOfTuples "^(string_of_setOfTuples s)^".\n"); *)
    let result=List.fold_left (fun sofar_new_setOfTuples curr_cp -> union_CP_setOfTuples (apply_transition_CP curr_cp comp was is) sofar_new_setOfTuples) [] s
    in (* Debug.print_debug "CartesianProduct.apply_transition_setOfTuples finished.\n"; *)
      result


  (* compute the number of elements in a cartesian product *)
  let rec size_of_cp (cp:t) :Big_int.big_int=
    let len=Array.length cp in if len<=0 then Big_int.zero_big_int else 
	let s=Big_int.big_int_of_int (SetOfComponentType.cardinal cp.(0)) in
	  if len=1 then s else Big_int.mult_big_int s (size_of_cp (Array.sub cp 1 (len-1)))

  (* extract_tuples_with_fixed_coords i s sot returns those tuples from sot 
     whose ith coordinate belongs to s. *)
  let rec extract_tuples_with_fixed_coords i s sot = 
    match sot with
	[] -> []
      | h::t -> let sot'=extract_tuples_with_fixed_coords i s t and
	    intr=SetOfComponentType.inter h.(i) s in
	  if SetOfComponentType.is_empty intr then sot' else 
	    union_CP_setOfTuples (Array.concat [(Array.sub h 0 i);(Array.make 1 intr);(Array.sub h (i+1) ((Array.length h)-i-1))]) sot'
		
  exception ExceptionSetNotFound
    
  (* exception_set_for_cp c b f where c and f are products and b is a set of 
     tuples finds some subset e of b such that the intersection of f with the 
     cartesian abstraction of (c union (b\e)) is empty. It assumes that the 
     intersection of f with (c union b) is empty. Possibly small set (wrt. 
     inclusion) to be returned is future work: it requires computing the 
     cardinality of a set of tuples. Raises Not_found if f has a nonempty
     intersection with c. *)
  let exception_set_for_cp c b fcp =
    (*Debug.print_debug ("CartesianProduct.exception_set_for_cp started with c="^(string_of_CP c)^" and b="^(string_of_setOfTuples b)^" and f="^(string_of_CP fcp)^".\n");*)
    let result =
      let len=Array.length fcp in if len<=0 then empty_setOfTuples 
	else 
	  let rec exception_set_for_cp_starting_from_index i =
	    if i>=len then raise ExceptionSetNotFound
	    else if (is_empty c) || (SetOfComponentType.is_empty (SetOfComponentType.inter fcp.(i) c.(i))) then extract_tuples_with_fixed_coords i fcp.(i) b 
	    else exception_set_for_cp_starting_from_index (i+1)
	  in exception_set_for_cp_starting_from_index 0
    in 
      (*Debug.print_debug ("CartesianProduct.exception_set_for_cp finished with result "^(string_of_setOfTuples result)^".\n");*)
      result


  (* eager_exception_set_for_cp c b f where c and f are products and b is a set of 
     tuples finds some subset e of b such that the intersection of f with the 
     cartesian abstraction of (c union (b\e)) is empty. It assumes that the 
     intersection of f with (c union b) is empty. A somewhat larger set (wrt. 
     inclusion) is returned than from exception_set_for_cp. *)
  let eager_exception_set_for_cp c b fcp =
(*    b *)
    let len=Array.length fcp in if len<=0 then empty_setOfTuples 
      else 
	let rec eager_exception_set_for_cp_starting_from_index i =
	  if i>=len then empty_setOfTuples
	  else if (is_empty c) || (SetOfComponentType.is_empty (SetOfComponentType.inter fcp.(i) c.(i))) then 
	    union_setOfTuples_setOfTuples
	      (extract_tuples_with_fixed_coords i fcp.(i) b)
	      (eager_exception_set_for_cp_starting_from_index (i+1))
	  else eager_exception_set_for_cp_starting_from_index (i+1)
	in eager_exception_set_for_cp_starting_from_index 0


  (* exception_set_for_setOfTuples c b f where c is a product and b and f are
     sets of tuples finds some subset e of b such that the intersection of f 
     with the cartesian abstraction of (c union (b\e)) is empty. It assumes 
     that the intersection of f with (c union b) is empty. *)
  let rec exception_set_for_setOfTuples c b f = 
    (*
    Debug.print_debug ("CartesianProduct.exception_set_for_setOfTuples started on c="^(string_of_CP c)^" and b="^(string_of_setOfTuples b)^" and f="^(string_of_setOfTuples f)^".\n"); *)
    let result= match f with
	[] -> []
      | h::t -> union_setOfTuples_setOfTuples (exception_set_for_cp c b h) (exception_set_for_setOfTuples c b t)
    in (* Debug.print_debug ("CartesianProduct.exception_set_for_setOfTuples finished with result "^(string_of_setOfTuples result)^".\n"); *)
      result

  (* eager_exception_set_for_setOfTuples c b f where c is a product and b and f are
     sets of tuples finds some subset e of b such that the intersection of f 
     with the cartesian abstraction of (c union (b\e)) is empty. It assumes 
     that the intersection of f with (c union b) is empty. Creates a somewhat larger set than
     exception_set_for_setOfTuples.
  *)
  let rec eager_exception_set_for_setOfTuples c b f = 
    (*
    Debug.print_debug ("CartesianProduct.eager_exception_set_for_setOfTuples started on c="^(string_of_CP c)^" and b="^(string_of_setOfTuples b)^" and f="^(string_of_setOfTuples f)^".\n"); *)
    let result= match f with
	[] -> []
      | h::t -> union_setOfTuples_setOfTuples (eager_exception_set_for_cp c b h) (eager_exception_set_for_setOfTuples c b t)
    in (* Debug.print_debug ("CartesianProduct.eager_exception_set_for_setOfTuples finished with result "^(string_of_setOfTuples result)^".\n"); *)
      result


  (* Creates a set of tuples { x1...xn | exists i: xi in inside and 
     for all j unequal i we have xj in outside} *)
  let mutex_setOfTuples (inside:Set.Make(Ord).t) (outside:Set.Make(Ord).t) n =
    let rec mutex_aux (i:int) (sofar:setOfTuples):setOfTuples =
      if i<0 then sofar else
	mutex_aux (i-1) (union_CP_setOfTuples (mutex_CP inside outside i n) sofar)
    in mutex_aux (n-1) empty_setOfTuples

  (* Creates a set of tuples { x1...xn | exists i unequal j: xi in inside and 
     xj in inside and for all k unequal i,j we have xk in all} *)
  let no_mutex_setOfTuples inside all n =
    let rec no_mutex_aux i j (sofar:setOfTuples):setOfTuples =
      if i<=0 then sofar else
	if j<0 then no_mutex_aux (i-1) (i-2) sofar else
	  no_mutex_aux i (j-1) 
	    (union_CP_setOfTuples (no_mutex_CP inside all i j n) sofar)
    in no_mutex_aux (n-1) (n-2) empty_setOfTuples

end
