(** License: GPLv3. *)

(** A Cartesian product of a user-given fixed dimension over a single user-given ordered type. Implementation: array of sets. For each underlying type, the empty Cartesian product has a unique representation, in which all the components are empty. *)

module type OrderedTypeWithPrinting = sig
  type t
  val compare: t->t->int
  val to_string: t->string
end

module type C = sig
  type componentType 
    (** The type of the components *)

  type setOfComponentType
    (** The type of a set of components *)

  type t
    (** The type of the Cartesian products *)

  val empty: t
    (** [empty n] is the empty Cartesian product of dimension [n] *)

  val is_empty: t->bool
    (** Tests whether a Cartesian product is empty or not. *)

  val make: int -> setOfComponentType -> t
    (** [make n s] returns a new empty CartesianProduct of size n, whose each
	component is initialized by the set s. *)

  val join: t-> t-> t
    (** [join c d] returns the smallest Cartesian product containing both [c] and [d]. Raises Invalid_argument if neither [c] nor [d] is empty but the dimensions of [c] and [d] are unequal. *)

  val inter: t-> t-> t
    (** [inter c d] returns a new cartesian product, whose components are the intersection of the corresponding components of [c] and [d]. Returns an empty set if the dimensions of [c] and [d] are unequal. *)

  val compare: t-> t-> int
    (** Total ordering between cartesian products. [compare s1 s2] returns -1 if [s1] is smaller than [s2], +1 if [s1] is larger than [s2] and 0 if both are equal. *)

  val subset: t-> t-> bool
    (** [subset c d] tests whether the set [c] is a subset of the set [d]. *)

  val equal: t-> t-> bool
    (** [equal s1 s2] tests whether the Cartesian products [s1] and [s2] are
	equal as sets, that is, contain same elements. *)

  val string_of_CP: t-> string
    (** Creates a string representation of a cartesian product *)

  type setOfTuples
    (** An object of the type [setOfTuples] represents a set of tuples, represented as a set of cartesian products. *)

  val empty_setOfTuples: setOfTuples
    (** The empty set of tuples *)

  val is_empty_setOfTuples: setOfTuples ->bool
    (** Check whether a given set of tuples is empty or not *)

  val singleton_setOfTuples: componentType array -> setOfTuples
    (** Create a set of tuples containing only one tuple *)

  val choose_tuple_from_setOfTuples: setOfTuples -> componentType array
    (** Chooses some arbitrary tuple from a set of tuples or throws
	Not_found if no such exists. *)

  val union_CP_setOfTuples: t-> setOfTuples-> setOfTuples
    (** create a set of tuples representing the inion of a cartesian product 
	and another set of tuples. *)

  val add_tuple_to_setOfTuples: componentType array ->setOfTuples->setOfTuples
    (** Adds a tuple to a set of tuples *)
   
  val union_setOfTuples_setOfTuples: setOfTuples->setOfTuples->setOfTuples
    (** Returns the union of two sets of tuples *)

  val inter_setOfTuples_setOfTuples: setOfTuples->setOfTuples->setOfTuples
    (** Returns the intersection of two sets of tuples *)

  val subset_setOfTuples_setOfTuples: setOfTuples->setOfTuples->bool
    (** Returns true iff the first argument is a subset of the second. *)

  val diff_setOfTuples_setOfTuples: setOfTuples->setOfTuples->setOfTuples
    (** [diff_setOfTuples_setOfTuples sot1 sot2] computes sot1\sot2 *)

  val cartabs_diff_setOfTuples_setOfTuples: setOfTuples->setOfTuples->(t*setOfTuples)
    (** [cartabs_diff_setOfTuples_setOfTuples sot1 sot2] computes the 
	smallest cartesian product containing sot1\sot2 and returns it in the 
	first component. The second component contains a more saturated 
	version of sot2. *)

  val join_CP_cartabs_diff_setOfTuples_setOfTuples: t->setOfTuples->setOfTuples->(t*setOfTuples)
    (** [join_CP_cartabs_diff_setOfTuples_setOfTuples c sot1 sot2] computes 
	the join of c with the smallest cartesian product containing sot1\sot2 
	and returns it in the first component. The second component contains a 
	more saturated version of sot2. *)

  val cartabs_setOfTuples:setOfTuples->t
    (** compute the cartesian abstraction of a set of tuples. *)

  val setOfTuples_from_CP: t->setOfTuples
    (** Creates a set of tuples from a cartesian product. *)

  val apply_transition_setOfTuples: setOfTuples -> int -> componentType -> componentType -> setOfTuples
  (** [apply_transition_setOfTuples s comp was is] applies the transition 
     "was->is" to the component comp of each tuple in s and returnes the 
     obtained set. *)

  exception ExceptionSetNotFound
  (** Gets thrown if a precondition if finding exception set is violated. *)

  val exception_set_for_setOfTuples: t->setOfTuples->setOfTuples->setOfTuples
  (** [exception_set_for_setOfTuples c b f] where c is a product and b and f 
      are sets of tuples finds some subset e of b such that the intersection 
      of f with the cartesian abstraction of (c union (b\e)) is empty. 
      It assumes that the intersection of f with (c union b) is empty. Raises 
      Not_found if f has a nonempty intersection with c. *)

  val eager_exception_set_for_setOfTuples: t->setOfTuples->setOfTuples->setOfTuples
  (** [eager_exception_set_for_setOfTuples c b f] where c is a product and b and f 
      are sets of tuples finds some subset e of b such that the intersection 
      of f with the cartesian abstraction of (c union (b\e)) is empty. 
      It assumes that the intersection of f with (c union b) is empty. Creates a somewhat larger set
      than exception_set_for_setOfTuples.
  *)


  val mutex_setOfTuples: (setOfComponentType)->(setOfComponentType)->int->setOfTuples
    (** [mutex_setOfTuples inside outside n] returns the set {x1...xn | 
	exists i: xi in inside and for all j unequal i: xj in outside }. *)

  val no_mutex_setOfTuples: setOfComponentType->setOfComponentType->int->setOfTuples
  (** [mutex_setOfTuples inside all n] creates the set of tuples 
      { x1...xn | exists i unequal j: xi in inside 
      and xj in inside and for all k unequal i,j we have xk in all}. *)

  val string_of_setOfTuples:setOfTuples->string
    (** creates a string representation of a tuple set. *)

  val string_of_setOfTuples_possibly_with_brackets:setOfTuples->string
    (** creates a string representation of a tuple set, when used for 
	multiplication. *)
end

module Make (Ord : OrderedTypeWithPrinting) : (C with type componentType = Ord.t and type setOfComponentType=Set.Make(Ord).t)

