(***********************************************************************)
(*                                                                     *)
(*                           Objective Caml                            *)
(*                                                                     *)
(*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         *)
(*                                                                     *)
(*  Copyright 1996 Institut National de Recherche en Informatique et   *)
(*  en Automatique.  All rights reserved.  This file is distributed    *)
(*  under the terms of the GNU Library General Public License, with    *)
(*  the special exception on linking described in file ../LICENSE.     *)
(*                                                                     *)
(*  Alexander Malkis. Added function max_key.                          *)
(***********************************************************************)

(* $Id: map.mli,v 1.33 2005/10/25 18:34:07 doligez Exp $ *)
(* $Id: extmap.mli,v 1.34 2008/03/07 20:59:37 alexmalk Exp $ *)

(** Association tables over ordered types.

   This module implements applicative association tables, also known as
   finite maps or dictionaries, given a total ordering function
   over the keys.
   All operations over maps are purely applicative (no side-effects).
   The implementation uses balanced binary trees, and therefore searching
   and insertion take time logarithmic in the size of the map.
*)

module type OrderedType =
  sig
    type t
      (** The type of the map keys. *)
    val compare : t -> t -> int
      (** A total ordering function over the keys.
          This is a two-argument function [f] such that
          [f e1 e2] is zero if the keys [e1] and [e2] are equal,
          [f e1 e2] is strictly negative if [e1] is smaller than [e2],
          and [f e1 e2] is strictly positive if [e1] is greater than [e2].
          Example: a suitable ordering function is the generic structural
          comparison function {!Pervasives.compare}. *)
  end
(** Input signature of the functor {!Map.Make}. *)

module type S =
  sig
    type key
    (** The type of the map keys. *)

    type (+'a) t
    (** The type of maps from type [key] to type ['a]. *)

    val empty: 'a t
    (** The empty map. *)

    val is_empty: 'a t -> bool
    (** Test whether a map is empty or not. *)

    val is_singleton : 'a t -> bool
    (** Test whether a map contains exactly one binding or not. *)

    val add: key -> 'a -> 'a t -> 'a t
    (** [add x y m] returns a map containing the same bindings as
       [m], plus a binding of [x] to [y]. If [x] was already bound
       in [m], its previous binding disappears. *)

    val find: key -> 'a t -> 'a
    (** [find x m] returns the current binding of [x] in [m],
       or raises [Not_found] if no such binding exists. *)

    val remove: key -> 'a t -> 'a t
    (** [remove x m] returns a map containing the same bindings as
       [m], except for [x] which is unbound in the returned map. *)

    val mem: key -> 'a t -> bool
    (** [mem x m] returns [true] if [m] contains a binding for [x],
       and [false] otherwise. *)

    val iter: (key -> 'a -> unit) -> 'a t -> unit
    (** [iter f m] applies [f] to all bindings in map [m].
       [f] receives the key as first argument, and the associated value
       as second argument.  The bindings are passed to [f] in increasing
       order with respect to the ordering over the type of the keys.
       Only current bindings are presented to [f]:
       bindings hidden by more recent bindings are not passed to [f]. *)

    val map: ('a -> 'b) -> 'a t -> 'b t
    (** [map f m] returns a map with same domain as [m], where the
       associated value [a] of all bindings of [m] has been
       replaced by the result of the application of [f] to [a].
       The bindings are passed to [f] in increasing order
       with respect to the ordering over the type of the keys. *)

    val mapi: (key -> 'a -> 'b) -> 'a t -> 'b t
    (** Same as {!Map.S.map}, but the function receives as arguments both the
       key and the associated value for each binding of the map. *)

    val fold: (key -> 'a -> 'b -> 'b) -> 'a t -> 'b -> 'b
    (** [fold f m a] computes [(f kN dN ... (f k1 d1 a)...)],
       where [k1 ... kN] are the keys of all bindings in [m]
       (in increasing order), and [d1 ... dN] are the associated data. *)

    val compare: ('a -> 'a -> int) -> 'a t -> 'a t -> int
    (** Total ordering between maps.  The first argument is a total ordering
        used to compare data associated with equal keys in the two maps. *)

    val equal: ('a -> 'a -> bool) -> 'a t -> 'a t -> bool
    (** [equal cmp m1 m2] tests whether the maps [m1] and [m2] are
       equal, that is, contain equal keys and associate them with
       equal data.  [cmp] is the equality predicate used to compare
       the data associated with the keys. *)

    val min_binding: 'a t->key*'a
      (** Returns the minimal binding of a map. *)

    val max_key: 'a t->key 
      (** Return the greatest key of the map (with respect to the 
       [Ord.compare_left] ordering), or raise Not_found if the map is empty. *)

    val for_all: 'a t->(key->'a->bool)->bool
      (** [for_all m f] checks whether all pairs [(k,v)] in [m] satisfy [f], 
	  i.e. [f k v] is true. *)

    val exists: 'a t->(key->'a->bool)->(key*'a)
      (** [exists m f] checks whether there is any binding [(k,v)] in [m] for 
	  which [f k v] is true and returns it. If no such binding exists,
	  Not_found is raised. *)

    val any_binding: 'a t ->(key*'a)
      (** Returns any key from the map in constant time. Raises Not_found if
	  the map is empty. *)

    val intersect_apply: 'a t -> 'a t -> ('a -> 'a -> ('a option)) -> 'a t
      (** [intersect_apply m1 m2 f] creates a new map that contains 
	  {(k,v)| ex. v1,v2: (k,v1)in m1, (k,v2)in m2, Some v = f v1 v2}*)

    val diff_apply: ('a t) -> ('a t) -> ('a -> 'a -> ('a option)) -> ('a t)
      (** [diff_apply s1 s2 f] computes a difference between maps s1 and s2,
	  applying each time a function f on matching keys. More precisely,
	  it computes 
	  {(k,v) | (k,v)in s1 and k is not a key of s2
	          or
	           (k,v1)in s1 and (k,v2)in s2 and f v1 v2 = Some v
	  }
      *)

    val diff_apply_change: ('a t)->('a t)->('a -> 'a option -> ('b option *'a option))->(('b t)*('a t))
      (** Computes a difference between maps s1 and s2. If a key occurs only in
	  s1, the binding is left "as is" in the difference. If a key k occurs 
	  both in s1 and s2 with data d1 and d2, and if fst(f d1 d2) = Some d, 
	  then (k,d) is also a binding of the difference. If a key does not 
	  occur in s1 or if a key occurs in s1 with data d1 and in s2 with data
	  d2, but fst(f d1 d2)=None, then the key doesn't belong to the 
	  difference. The difference is returned as the first component of the 
	  function. Additionally, a rebuild of the map s2 is returned in the 
	  second component. A rebuild of s2 is a map s2' on the same set of 
	  keys so that for all keys k of s2 (or s2') we have: whenever (k,d2) 
	  is in s2 and (k,d2') in s2', then either d2=d2' or there is some d1 
	  with (k,d1) in s1 and snd(f d1 (Some d2))=Some d2'; moreover, if 
	  snd(f d1 (Some d2))=None, then d2=d2'. *)

    val union: ('a ->'a ->'a)->'a t->'a t->'a t
      (** [union f s1 s2] creates a map containing 
	  {(k,d)| (k,d1)in s1 and k is not a key of s2 and d1=d} 
	  union
	  {(k,d)| (k,d2)in s2 and k is not a key of s1 and d2=d} 
	  union
	  {(k,d) | (k,d1)in s1 and (k,d2)in s2 and f d1 d2 = d}. *)

    val union_apply3: ('a option -> 'b ->'c option ->(('a option)*('c option)))->('a t)->('b t)->('c t)->(('a t)*('c t))
      (** [union_apply3 f s1 s2 s3] creates the map
	  {(k,d)| (k,d1)in s1 and k is not a key of s2 and d1 = d} 
	  union
	  {(k,d)| (k,d2)in s2 and k is not a key of s1 and 
	  ((k,d3) is in s3 and fst(f None d2 (Some d3)) = Some d)
	  or 
	  (k is not a key of s3 and fst(f None d2 None) = Some d)}
	  }
	  union
	  {(k,d)| (k,d1)in s1 and (k,d2)in s2 and 
	  ((k,d3) is in s3 and fst (f (Some d1) d2 (Some d3)) = Some d)
	  or 
	  (k is not a key of s3 and fst(f (Some d1) d2 None) = Some d)}. 
	  This map is returned in the first component.
	  Additionally, a rebuild of the map s3 is returned in the second 
	  component. A rebuild of s3 is a map s3' on the same set of keys so 
	  that for all keys k of s3 (or s3') we have: whenever (k,d3) is in s3 
	  and (k,d3') in s3', then either d3=d3' or there is some d2 with 
	  (k,d2) in s2 and snd(f _ d2 (Some d3))=Some d3'. *)

    val subset_apply: ('a->'a->bool)->'a t->'a t->bool
    (** subset function checks whether each key of s1 is a key of s2 and 
	whenever (k,d1) is a binding of s1 and (k,d2) of s2, f d1 d2 holds *)  

    val cardinal: 'a t -> int
    (** [cardinal m] returns the number of bindings of m *)

  end
(** Output signature of the functor {!Map.Make}. *)

module Make (Ord : OrderedType) : S with type key = Ord.t
(** Functor building an implementation of the map structure
   given a totally ordered type. *)
