(* Written by Alexander Malkis, 2009. Licence: GPLv3 *)

type glob = int
type loc = int
type thread_state = glob * loc
type prog_state = glob * (loc array)
type thread_transition = thread_state * int * thread_state

let rec foldl2_array (f: 'a -> ('b*'b) -> 'a) (init:'a) ((bar1:('b array)),(bar2
:('b array))) :'a =
  let
    len = Array.length bar1 and
    len2 = Array.length bar2
  in
    if len <> len2 then raise (Invalid_argument "foldl2_array expects arrays of
equal size") else
    if len=0 then init else
      f (foldl2_array f init (Array.sub bar1 1 (len - 1), Array.sub bar2 1 (len-
1))) (bar1.(0),bar2.(0))


module PairOfOrderedGlobStr = 
  struct
    type left_type=glob
    type right_type=string
    let compare_left (l:left_type) (l':left_type) = Pervasives.compare l l'
    let compare_right (r:right_type) (r':right_type) = String.compare r r'
  end
  
module PairOfOrderedLocStr = 
  struct
    type left_type=loc
    type right_type=string
    let compare_left (l:left_type) (l':left_type) = Pervasives.compare l l'
    let compare_right (r:right_type) (r':right_type) = String.compare r r'
  end

module BijectiveMapGlobStr = BijectiveMap.Make (PairOfOrderedGlobStr)
type bijectiveMapGlobStr=BijectiveMapGlobStr.t
module BijectiveMapLocStr = BijectiveMap.Make (PairOfOrderedLocStr)
type bijectiveMapLocStr=BijectiveMapLocStr.t

(* module for defining abstract and concrete elements *)
module SL = struct
  type shared = glob*string
  type component = loc*string
  let compare_shared (g1,s1) (g2,s2) = Pervasives.compare g1 g2
  let compare_component (c1,s1) (c2,s2) = Pervasives.compare c1 c2
  let string_of_shared =snd
  let string_of_component =snd
end


module AC = AbstractConcreteElement.Make(SL)

(* Threads werden durchnummeriert von 0 bis no_of_threads-1, globale Zustnde von 0 bis no_of_globals-1, lokale Zustnde von 0 bis no_of_locals-1 *)
type verificationTask = {no_of_globals:int; glob_to_str:bijectiveMapGlobStr; no_of_locals:int; loc_to_str:bijectiveMapLocStr; no_of_threads:int; all_transitions:AC.transitionSet; init:AC.setOfStates; error:AC.setOfStates}


(*
 * String prefix.
 *)
let rec is_string_prefix_aux s1 s2 i len =
   i = len || (String.unsafe_get s1 i = String.unsafe_get s2 i && is_string_prefix_aux s1 s2 (i + 1) len)

let is_string_prefix s1 s2 =
   let len1 = String.length s1 in
   let len2 = String.length s2 in
      len1 <= len2 && is_string_prefix_aux s1 s2 0 len1


(*
 * Split a string str into a list of substrings.
 * The string is split on any character in delims.  Quotations
 * are not split.
 *
 * Empty substrings are _not_ returned as empty strings in the list.
 * For example:
 *   split ".-" "foo.bar--ba??z"
 * returns
 *   ["foo"; "bar"; "ba??z"]
 *)
let tokens_fold f x quotes delims str =
   let strlen = String.length str in
   (* Skip white space *)
   let rec skip_split pos =
      if pos = strlen then strlen
      else
         let c = str.[pos] in
            if String.contains delims c then skip_split (succ pos)
            else pos
   in
   (* Find the next split index. *)
   let rec next_split pos =
      if pos = strlen then strlen
      else
         let c = str.[pos] in
            if String.contains delims c then pos
            else if String.contains quotes c then next_quote (succ pos)
            else next_split (succ pos)

   and next_quote pos =
      if pos = strlen then strlen
      else
         let c = str.[pos] in
            if String.contains quotes c then next_split (succ pos)
            else if c = '\\' && pos < pred strlen then next_quote (pos + 2)
            else next_quote (succ pos)
   in
   (* Build the list *)
   let rec str_split x pos =
      if pos = strlen then x
      else
         let pos_end = next_split pos in
         let x = f x str pos (pos_end - pos) in
            str_split x (skip_split pos_end)
   in str_split x (skip_split 0)

(*
 * Default token processor.
 *)
let tokens quotes delims str =
   let l =
      tokens_fold (fun l s off len ->
            String.sub s off len :: l) [] quotes delims str
   in List.rev l

let tokens_std (s:string): string list = tokens "\"'" " \t\r\n\012" s




(*Reads a verification taks from the standard input*)
let read_verificationTask () : verificationTask = 
  let rec read_verifTask_accum (read_until_now:verificationTask):verificationTask = 
    try
      let input_line = read_line ()
      in let ts = tokens_std input_line
      in let new_verif_task=
	  if ((List.length ts)=0) or (
	    let ts_head=List.hd ts
	    in (((String.length ts_head)<>0)&&('#'=(String.get ts_head 0)))
	  ) 
	  then read_until_now
	  else 
	    let rec get_local_states list_remainder =
	      (* Convert the list of locals into an array *)
	      if ((List.length list_remainder)=0) then ((Array.make 0 (0,"")),(read_until_now.loc_to_str),(read_until_now.no_of_locals)) else 
		let local_str=List.hd list_remainder
		and (remainder_array,new_loc_to_str_map,no_of_locals)=get_local_states (List.tl list_remainder)
		in let (local_state,new_local_state_needed)= 
		    try (BijectiveMapLocStr.find_right local_str new_loc_to_str_map,false)
		    with Not_found  -> ((try ((BijectiveMapLocStr.max_left new_loc_to_str_map)+1) with Not_found -> 0),true)
		in ((Array.append (Array.make 1 (local_state,local_str)) remainder_array),(if new_local_state_needed then BijectiveMapLocStr.add local_state local_str new_loc_to_str_map else new_loc_to_str_map), (if new_local_state_needed then no_of_locals+1 else no_of_locals))
            in if (String.compare "T" (List.hd ts))=0 then 
		(* Input a new transition *)
		(if (List.length ts)<>6 then raise (Invalid_argument ("There is a bad inpit line that starts with T:"^input_line^"\nFormat:\nT <thread_number> <global> <local> <global'> <local'>\n")) else
		   let thread_no_str=List.hd (List.tl ts)
		   and global_str=List.hd (List.tl (List.tl ts))
		   and local_str=List.hd (List.tl (List.tl (List.tl ts)))
		   and global_next_str=List.hd (List.tl (List.tl (List.tl (List.tl ts))))
		   and local_next_str=List.hd (List.tl (List.tl (List.tl (List.tl (List.tl ts)))))
		   in let thread_no=
		       try int_of_string thread_no_str
		       with Failure s -> raise (Invalid_argument ("The input line\n"^input_line^"\nspecifies a non-numeric thread number. Format:\nT <thread_number> <global> <local> <global'> <local'>\n"))
		      and (global_state,new_global_state_needed)=
		       try (BijectiveMapGlobStr.find_right global_str read_until_now.glob_to_str,false)
		       with Not_found  -> ((try ((BijectiveMapGlobStr.max_left read_until_now.glob_to_str)+1) with Not_found -> 0),true)
		   in let new_glob_to_str=if new_global_state_needed then BijectiveMapGlobStr.add global_state global_str read_until_now.glob_to_str else read_until_now.glob_to_str
		      and (local_state,new_local_state_needed)= 
		       try (BijectiveMapLocStr.find_right local_str read_until_now.loc_to_str,false)
		       with Not_found  -> ((try ((BijectiveMapLocStr.max_left read_until_now.loc_to_str)+1) with Not_found -> 0),true)
		   in let new_loc_to_str = if new_local_state_needed then BijectiveMapLocStr.add local_state local_str read_until_now.loc_to_str else read_until_now.loc_to_str
		      and (global_next_state,new_global_next_state_needed)=
		       try (BijectiveMapGlobStr.find_right global_next_str new_glob_to_str,false)
		       with Not_found  -> ((try ((BijectiveMapGlobStr.max_left new_glob_to_str)+1) with Not_found -> 0),true)
		   in let new_new_glob_to_str=if new_global_next_state_needed then BijectiveMapGlobStr.add global_next_state global_next_str new_glob_to_str else new_glob_to_str
		      and (local_next_state,new_local_next_state_needed)= 
		       try (BijectiveMapLocStr.find_right local_next_str new_loc_to_str,false)
		       with Not_found  -> ((try ((BijectiveMapLocStr.max_left new_loc_to_str)+1) with Not_found -> 0),true)
		   in let new_new_loc_to_str=if new_local_next_state_needed then BijectiveMapLocStr.add local_next_state local_next_str new_loc_to_str else new_loc_to_str
		      and new_transition=(AC.single_transition (global_state,global_str) (local_state,local_str) thread_no (global_next_state,global_next_str) (local_next_state,local_next_str))
		   in let new_transition_set=AC.add_transition_transitionSet new_transition read_until_now.all_transitions
		   in {no_of_globals=read_until_now.no_of_globals+(if new_global_state_needed then (if new_global_next_state_needed then 2 else 1) else (if new_global_next_state_needed then 1 else 0)); glob_to_str=new_new_glob_to_str; no_of_locals=read_until_now.no_of_locals +(if new_local_state_needed && new_local_next_state_needed then 2 else (if (new_local_state_needed || new_local_next_state_needed) then 1 else 0)); loc_to_str=new_new_loc_to_str; no_of_threads=if read_until_now.no_of_threads<=thread_no then thread_no+1 else  read_until_now.no_of_threads; all_transitions=new_transition_set; init=read_until_now.init; error=read_until_now.error})
	      else if ((String.compare "Init" (List.hd ts))=0) then 
		(* Input a new initial state. *)
		(let number_of_threads=(List.length ts)-2
		 in if number_of_threads<=0 then raise (Invalid_argument ("There is a bad input line starting with Init, namely no threads are given:\n"^input_line^"\nFormat:\nInit <global> <local_1> ... <local_n>\n")) else if number_of_threads<read_until_now.no_of_threads then raise (Invalid_argument ("There is a bad input line starting with Init:\n"^input_line^"\nThe number of threads that were read until now is "^(string_of_int read_until_now.no_of_threads)^", which exceeds the number of threads specified by this initial state.\n")) else if (not (AC.is_empty_setOfStates read_until_now.init)) && ((AC.state_dimension (AC.choose_state_from_setOfStates read_until_now.init))<>number_of_threads) then raise (Invalid_argument ("There is a bad input line starting with Init:\n"^input_line^"\nA previously declared initial state had "^(string_of_int (AC.state_dimension (AC.choose_state_from_setOfStates read_until_now.init)))^" threads.\n")) else if (not (AC.is_empty_setOfStates read_until_now.error)) && ((AC.state_dimension (AC.choose_state_from_setOfStates read_until_now.error))<>number_of_threads) then raise (Invalid_argument ("There is a bad input line starting with Init:\n"^input_line^"\nA previously declared error state had "^(string_of_int (AC.state_dimension (AC.choose_state_from_setOfStates read_until_now.error)))^" threads.\n")) else 
		     let global_str=List.hd (List.tl ts)
		     and locals_list=List.tl (List.tl ts)
		     in let (global_state,new_global_state_needed)=
			 try (BijectiveMapGlobStr.find_right global_str read_until_now.glob_to_str,false)
			 with Not_found  -> ((try ((BijectiveMapGlobStr.max_left read_until_now.glob_to_str)+1) with Not_found -> 0),true)
		     in let new_glob_to_str=if new_global_state_needed then BijectiveMapGlobStr.add global_state global_str read_until_now.glob_to_str else read_until_now.glob_to_str
			and (locals_array,new_loc_to_str,new_no_of_locals)=get_local_states locals_list
		     in {no_of_globals=read_until_now.no_of_globals+(if new_global_state_needed then 1 else 0); glob_to_str = new_glob_to_str; no_of_locals=new_no_of_locals; loc_to_str=new_loc_to_str; no_of_threads=number_of_threads; all_transitions=read_until_now.all_transitions; init=AC.add_state_setOfStates (AC.single_state (global_state,global_str) locals_array) read_until_now.init;error=read_until_now.error})
	      else if (String.compare "Error" (List.hd ts))=0 then 
		(* Input a new error state *)
		(let number_of_threads=(List.length ts)-2
		 in if number_of_threads<=0 then raise (Invalid_argument ("There is a bad input line starting with Error:\n"^input_line^"\nFormat:\nError <global> <local_1> ... <local_n>\n")) else if number_of_threads<read_until_now.no_of_threads then raise (Invalid_argument ("There is a bad input line starting with Error:\n"^input_line^"\nThe number of threads that were read until now is "^(string_of_int read_until_now.no_of_threads)^", which exceeds the number of threads specified by this initial state.\n")) else if (not (AC.is_empty_setOfStates read_until_now.error)) && ((AC.state_dimension (AC.choose_state_from_setOfStates read_until_now.error))<>number_of_threads) then raise (Invalid_argument ("There is a bad input line starting with Error:\n"^input_line^"\nA previously declared error state had "^(string_of_int (AC.state_dimension (AC.choose_state_from_setOfStates read_until_now.error)))^" threads.\n")) else if (not (AC.is_empty_setOfStates read_until_now.init)) && ((AC.state_dimension (AC.choose_state_from_setOfStates read_until_now.init))>number_of_threads) then raise (Invalid_argument ("There is a bad input line starting with Error:\n"^input_line^"\nA previously declared initial state had "^(string_of_int (AC.state_dimension (AC.choose_state_from_setOfStates read_until_now.init)))^" threads.\n")) else
		     let global_str=List.hd (List.tl ts)
		     and locals_list=List.tl (List.tl ts)
		     in let (global_state,new_global_state_needed)=
			 try (BijectiveMapGlobStr.find_right global_str read_until_now.glob_to_str,false)
			 with Not_found  -> ((try ((BijectiveMapGlobStr.max_left read_until_now.glob_to_str)+1) with Not_found -> 0),true)
		     in let new_glob_to_str = if new_global_state_needed then BijectiveMapGlobStr.add global_state global_str read_until_now.glob_to_str else read_until_now.glob_to_str
			and (locals_array,new_loc_to_str,new_no_of_locals)=get_local_states locals_list
		     in {no_of_globals=read_until_now.no_of_globals+(if new_global_state_needed then 1 else 0); glob_to_str = new_glob_to_str; no_of_locals=new_no_of_locals; loc_to_str=new_loc_to_str; no_of_threads=number_of_threads; all_transitions=read_until_now.all_transitions; init=read_until_now.init; error=AC.add_state_setOfStates (AC.single_state (global_state,global_str) locals_array) read_until_now.error})
	      else (raise (Invalid_argument ("In the input, some line starts with an unknown word "^(List.hd ts)^"\n")))
      in read_verifTask_accum new_verif_task
    with End_of_file -> read_until_now
  in read_verifTask_accum {no_of_globals=0; glob_to_str=BijectiveMapGlobStr.empty; no_of_locals=0; loc_to_str=BijectiveMapLocStr.empty; no_of_threads=0; all_transitions=AC.empty_transitionSet; init=AC.empty_setOfStates; error=AC.empty_setOfStates}
       

(* conversion to string maps *)

(* Converts a verification task into a string *)
let string_of_verificationTask (vt:verificationTask) =
  "Transitions:\n"^(AC.string_of_transitionSet vt.all_transitions)^"\nInitial states:\n"^(AC.string_of_setOfStates vt.init)^"\nError states:\n"^(AC.string_of_setOfStates vt.error)^"\n"


(* checking whether a list is empty *)
let is_list_empty = (function
    [] -> true
  | _ -> false)

(* Given a list of set of states (s_i)_{0\le i<m} and a transition relation 
post, compute inductively t_0=s_0 and t_{i+1}=(s_{i+1} intersection post(t_i))
*)
let relative_post_wrt_list (lst:AC.setOfStates list) (post:AC.transitionSet) =
  match lst with
      [] -> []
    | (h::tl) -> 
	let (last,tail)=List.fold_left (fun (last_sofar,sofar_prev) curr -> ((AC.intersection_setOfStates_setOfStates (AC.apply_transitionSet last_sofar post) curr),(last_sofar::sofar_prev))) (h,[]) tl
	in List.rev(last::tail)


(* what a model-checker returns *)
type model_checking_result=
    Safe of AC.setOfStates (* constructed invariant *)
  | Unsafe of AC.setOfStates list (* counterexample *)

(* The lazy model checking procedure. *)
let model_check (init:AC.setOfStates) (error:AC.setOfStates) (post:AC.transitionSet) =
(*
  let _=Debug.print_debug ("Transition relation = "^(AC.string_of_transitionSet post)^".\n") in
*)
  (*the predecessor relation*)
  let pre=AC.reverse_transitionSet post in
  
  (* stringify function *)
  let rec string_of_forward_trace (forward_trace:(AC.abstractElement*AC.setOfStates*AC.setOfStates*AC.setOfStates) list) :string =
    ("["^(
       let lst=(List.map (fun (abs_elem, conc_abs_elem, succ_conc_abs_elem,ex_set)-> "("^
 "absElem="^(AC.string_of_abstractElement abs_elem)^","^
 "concAbsElem="^(AC.string_of_setOfStates conc_abs_elem)^
","^ 
 "succ_conc_abs_elem="^(AC.string_of_setOfStates succ_conc_abs_elem)^","^ 
   "ex_set="^(AC.string_of_setOfStates ex_set)^
   ")") forward_trace) in
   String.concat ", " lst
   )^"]")
    
  (* lines 4-11 of the algorithm *)
  and forward_search (future_exception_sets:AC.setOfStates list) (ex_set_changed:bool) (sofar_forward:(AC.abstractElement*AC.setOfStates*AC.setOfStates*AC.setOfStates) list) =
(*
    Debug.print_debug ("forward_search started on sequence "^(string_of_forward_trace sofar_forward)^" and new exception set sequence "^(AC.string_of_trace future_exception_sets)^".\n");
*)
(*    Debug.print_debug ("Forward length "^(string_of_int (List.length sofar_forward))^", exception length "^(string_of_int (List.length future_exception_sets))^"\n");
*)
    match sofar_forward with
	[] -> forward_search [AC.empty_setOfStates] false [(AC.least_abstractElement, AC.empty_setOfStates, init, AC.empty_setOfStates)]
      | (last_abs_elem, _, succ_conc_last_abs_elem,_)::t -> 
	  let (next_exception_set,new_future_exception_sets,new_ex_set_changed)=
	    match future_exception_sets with
		[] -> raise (Failure "ModelChecker.forward_search accepted an empty exception set sequence\n")
	      | [h]  -> (h,future_exception_sets,false)
	      | (h::tl) -> (h,tl,true)
	  in let (next_abs_elem,optimized_next_exception_set)=AC.join_abstract_element_cartabs_setOfStates_minus_setOfStates last_abs_elem succ_conc_last_abs_elem next_exception_set
	  in let conc_next_abs_elem=AC.concretize_abstractElement_union_setOfStates next_abs_elem optimized_next_exception_set
	  in let succ_conc_next_abs_elem = AC.apply_transitionSet conc_next_abs_elem post
	  in let inter_with_error=AC.intersection_setOfStates_setOfStates conc_next_abs_elem error
	  in let is_still_safe=AC.is_empty_setOfStates inter_with_error in
	    if not is_still_safe then analyze_trace ((next_abs_elem,conc_next_abs_elem,succ_conc_next_abs_elem,next_exception_set)::sofar_forward) [inter_with_error]
	    else if (not ex_set_changed) && (AC.le_abstract_elements next_abs_elem last_abs_elem) then Safe conc_next_abs_elem
	    else 
	      forward_search new_future_exception_sets new_ex_set_changed ((next_abs_elem,conc_next_abs_elem,succ_conc_next_abs_elem,optimized_next_exception_set)::sofar_forward)

  (* lines 13-17 *)
  and analyze_trace forward_trace f_trace=
(*
    Debug.print_debug ("analyze_trace started on forward trace "^(string_of_forward_trace forward_trace)^" and backward trace "^(AC.string_of_trace f_trace)^".\n");
*)
    match (forward_trace,f_trace) with
	([],_) -> raise (Invalid_argument "An empty forward trace should be analyzed!\n")
      | (_,[]) -> raise (Invalid_argument "An empty backward trace is given!\n")
      | ([(_,_,_,_)],_) -> raise (Invalid_argument "Backward analysis tries to go beyond the initial states!")
      | (((_,_,_,_)::((prev_abs_elem,conc_prev_abs_elem,succ_conc_prev_abs_elem,ex_set)::forward_tail)),(f::backward_tail)) ->
	  if (is_list_empty forward_tail) && (not (AC.is_empty_setOfStates (AC.intersection_setOfStates_setOfStates f succ_conc_prev_abs_elem))) then Unsafe (relative_post_wrt_list f_trace post)
	  else
	    let pred_f_inside_conc_prev = 
	      AC.intersection_setOfStates_setOfStates 
(*
		(AC.apply_reversed_transitionSet f post) 
*)
		(AC.apply_transitionSet f pre)
		conc_prev_abs_elem 
	    in if AC.is_empty_setOfStates pred_f_inside_conc_prev then refine forward_trace f_trace
	      else analyze_trace ((prev_abs_elem,conc_prev_abs_elem,succ_conc_prev_abs_elem,ex_set)::forward_tail) (pred_f_inside_conc_prev::f_trace)

  (* generate a sequence of exceptions sets for the future. It is assumed that
     the head of sofar_generated_ex_set_list is attached to current_iterate. *)
  and generate_future_ex_set_list current_iterate f_trace_from_current sofar_generated_ex_set_list =
(*
    Debug.print_debug ("generate_future_ex_set_list started with current_iterate="^(AC.string_of_setOfStates current_iterate)^", f_trace_from_current="^(AC.string_of_trace f_trace_from_current)^", sofar_generated="^(AC.string_of_trace sofar_generated_ex_set_list)^"\n");
*)
    match f_trace_from_current with
	[] -> sofar_generated_ex_set_list
      | (f::remainder_f_trace) ->
	  let succ_curr_iterate= AC.apply_transitionSet current_iterate post in
	  let curr_iterate_abstraction=AC.abstract_setOfStates current_iterate 
	  in let delta_ex_set = try
	    AC.exception_set curr_iterate_abstraction succ_curr_iterate f
	  with AC.ExceptionSetNotFound ->
	    AC.exception_set AC.least_abstractElement succ_curr_iterate f
	  in let new_ex_set=AC.union_setOfStates_setOfStates (List.hd sofar_generated_ex_set_list) delta_ex_set in
	  let new_iterate=AC.union_setOfStates_setOfStates current_iterate succ_curr_iterate in
	    generate_future_ex_set_list new_iterate remainder_f_trace (new_ex_set::sofar_generated_ex_set_list)
	      
  (* lines 20-22 *)
  and refine forwards_trace f_trace =
(*
    Debug.print_debug ("refine started on forward trace "^(string_of_forward_trace forwards_trace)^" and bad region list "^(AC.string_of_trace f_trace)^".\n");
*)
(* Debug.print_debug ("refine started on forward trace of length "^(string_of_int (List.length forwards_trace))^".\n"); *)
    match forwards_trace with
	[] -> raise (Invalid_argument "An empty forwards trace is given!\n")
      | [(_,_,_,_)]->
	  raise (Invalid_argument "Refinement for the 0th position is asked!\n")
      | ((abs_elem,conc_abs_elem,succ_conc_abs_elem,ex_set)::((prev_abs_elem,conc_prev_abs_elem,succ_conc_prev_abs_elem,prev_ex_set)::t)) ->
	  match f_trace with
	      [] -> raise (Invalid_argument "Empty bad region trace!\n")
	    | (f::remainder_f) ->
		let delta_ex_set=AC.exception_set prev_abs_elem succ_conc_prev_abs_elem f in
		let new_ex_set=AC.union_setOfStates_setOfStates ex_set delta_ex_set in
		let (new_abs_elem,optimized_new_ex_set)=AC.join_abstract_element_cartabs_setOfStates_minus_setOfStates prev_abs_elem succ_conc_prev_abs_elem new_ex_set in
		let new_curr_iterate=AC.concretize_abstractElement_union_setOfStates new_abs_elem new_ex_set in
		let future_ex_set_list = List.rev (generate_future_ex_set_list new_curr_iterate remainder_f [optimized_new_ex_set]) in
		  forward_search future_ex_set_list true ((prev_abs_elem,conc_prev_abs_elem,succ_conc_prev_abs_elem,prev_ex_set)::t)

  in forward_search [AC.empty_setOfStates] false []


(* The eager model checking procedure. *)
let model_check_eager (init:AC.setOfStates) (error:AC.setOfStates) (post:AC.transitionSet) =
(*
  let _=Debug.print_debug ("Transition relation = "^(AC.string_of_transitionSet post)^".\n") in
*)
  (*the predecessor relation*)
  let pre=AC.reverse_transitionSet post in

  (* stringify function *)
  let rec string_of_forward_trace (forward_trace:((AC.abstractElement*AC.setOfStates*AC.setOfStates) list)) :string =
    ("["^(
       let lst=(List.map (fun (abs_elem, conc_abs_elem, succ_conc_abs_elem)-> "("^
 "absElem="^(AC.string_of_abstractElement abs_elem)^","^
 "concAbsElem="^(AC.string_of_setOfStates conc_abs_elem)^
","^ 
 "succ_conc_abs_elem="^(AC.string_of_setOfStates succ_conc_abs_elem)^
   ")") forward_trace) in
   String.concat ", " lst
   )^"]")
  in

  (* forward fixpoint iteration *)
  let rec forward_search (err:AC.setOfStates) (ex_set:AC.setOfStates) (sofar_forward:((AC.abstractElement*AC.setOfStates*AC.setOfStates) list)) =
(*
    Debug.print_debug ("forward_search started on error set "^(AC.string_of_setOfStates err)^", exception set "^(AC.string_of_setOfStates ex_set)^" and forward sequence "^(string_of_forward_trace sofar_forward)^".\n");
*)
    match sofar_forward with
	[] -> (forward_search err AC.empty_setOfStates [(AC.least_abstractElement, AC.empty_setOfStates,init)])
      | ((abs_elem,conc_abs_elem,succ_conc_abs_elem)::t) ->
	  let (next_abs_elem,optimized_ex_set)=(AC.join_abstract_element_cartabs_setOfStates_minus_setOfStates abs_elem succ_conc_abs_elem ex_set) in
	  let conc_next_abs_elem=AC.concretize_abstractElement_union_setOfStates next_abs_elem optimized_ex_set in
	  let succ_conc_next_abs_elem = AC.apply_transitionSet conc_next_abs_elem post in
	  let inter_with_error=AC.intersection_setOfStates_setOfStates conc_next_abs_elem err in
	    if AC.is_empty_setOfStates inter_with_error then
	      if (AC.le_abstract_elements next_abs_elem abs_elem) then 
		Safe conc_next_abs_elem
	      else forward_search err optimized_ex_set ((next_abs_elem,conc_next_abs_elem,succ_conc_next_abs_elem)::sofar_forward)
	    else analyze_trace err optimized_ex_set ((next_abs_elem,conc_next_abs_elem,succ_conc_next_abs_elem)::sofar_forward) [inter_with_error]

  (* backtracking *)
  and analyze_trace err ex_set forward_trace f_trace =
    match (forward_trace,f_trace) with
	([],_) -> raise (Invalid_argument "An empty forward trace should be analyzed!\n")
      | (_,[]) -> raise (Invalid_argument "An empty backward trace is given!\n")
      | ([(_,_,_)],_) -> raise (Invalid_argument "Backward analysis tries to go beyond the initial states!")
      | (((_,_,_)::((prev_abs_elem,conc_prev_abs_elem,succ_conc_prev_abs_elem)::forward_tail)),(f::backward_tail)) ->
	  let at_init = is_list_empty forward_tail in
	    if at_init && (not (AC.is_empty_setOfStates (AC.intersection_setOfStates_setOfStates f succ_conc_prev_abs_elem))) then Unsafe (relative_post_wrt_list f_trace post)
	    else if at_init then
	      let inter_f_with_ex_set = AC.intersection_setOfStates_setOfStates f ex_set in
		if AC.is_empty_setOfStates inter_f_with_ex_set then
		  refine err ex_set forward_trace f
		else
		  let new_err=AC.union_setOfStates_setOfStates err inter_f_with_ex_set in
		  let new_ex_set=AC.diff_setOfStates_setOfStates ex_set inter_f_with_ex_set in
		    forward_search new_err new_ex_set [(AC.least_abstractElement, AC.empty_setOfStates,init)]
	    else 
	      let pred_f_in_conc_prev=AC.intersection_setOfStates_setOfStates 
		(AC.apply_transitionSet f pre) conc_prev_abs_elem
	      in if AC.is_empty_setOfStates pred_f_in_conc_prev then
		  refine err ex_set forward_trace f
		else analyze_trace err ex_set ((prev_abs_elem,conc_prev_abs_elem,succ_conc_prev_abs_elem)::forward_tail) (pred_f_in_conc_prev::f_trace)

  (* compute a new exception set *)	  
  and refine err ex_set forward_trace f =
(*
    Debug.print_debug ("refine started on err="^(AC.string_of_setOfStates err)^", ex_set="^(AC.string_of_setOfStates ex_set)^", forward_trace="^(string_of_forward_trace forward_trace)^" and bad region "^(AC.string_of_setOfStates f)^".\n");
*)
      match forward_trace with
	  [] -> raise (Invalid_argument "An empty forwards trace is given!\n")
	| [(_,_,_)]->
	  raise (Invalid_argument "Refinement for the 0th position is asked!\n")
	| ((abs_elem,conc_abs_elem,succ_conc_abs_elem)::((prev_abs_elem,conc_prev_abs_elem,succ_conc_prev_abs_elem)::t)) ->
	    let delta_ex_set=AC.exception_set prev_abs_elem succ_conc_prev_abs_elem f in
	    let new_ex_set =AC.union_setOfStates_setOfStates ex_set delta_ex_set in
	      forward_search err new_ex_set [(AC.least_abstractElement, AC.empty_setOfStates,init)]

  in forward_search error AC.empty_setOfStates []
