(* Written by Alexander Malkis, 2009. Licence: GPLv3 *)

open ModelChecker


(* We will check the program of n threads, each thread consisting of m 
   critical sections of k locations. Each thread looks as follows:

Q(0):       acquire lock;
R(0,0):     nop;
...
R(0,k-2):   nop;
R(0,k-1):   release lock;
Q(1):       acquire lock;
R(1,0):     nop;
...
R(1,k-2):   nop;
R(1,k-1):   release lock;
...
Q(m-1):     acquire lock;
R(m-1,0):   nop;
...
R(m-1,k-2): nop;
R(m-1,k-1): <release lock; goto Q0;>

The numerical values of local states are
Q(j)=jk (0<=j<m)
R(j,r)=jk+r (0<=j<m, 0<=r<k)
*)

(* States. *)
module OrderedInt=struct
  type t=int
  let compare=Pervasives.compare
end

module MapFromInts=ExtMap.Make(OrderedInt)

(* Creates a map from naturals to states. *)
let create_states m k =
  let rec create_states_aux j r sofar =
    if j<0 then sofar
    else if r<0 then create_states_aux (j-1) k sofar
    else
      let new_state=(j*(k+1)+r,
		     if r=0 then ("Q("^(string_of_int j)^")")
		     else ("R("^(string_of_int j)^","^(string_of_int (r-1))^")")
		    )
      in create_states_aux j (r-1) (MapFromInts.add (j*(k+1)+r) new_state sofar)
  in create_states_aux (m-1) k MapFromInts.empty
	 
let unlocked=(0,"0")
let locked=(1,"1")
    
module SetOfComponent=Set.Make(AbstractConcreteElement.MakeOrderedLocal(SL))
module SetOfShared=Set.Make(AbstractConcreteElement.MakeOrderedShared(SL))

(* Creates the transition set going out from the state j(k+1)+r of thread i *)
let create_single_transition local_map m k i j r=
  let loc_source=MapFromInts.find (j*(k+1)+r) local_map in
  let loc_target=MapFromInts.find ((j*(k+1)+r+1) mod (m*(k+1))) local_map in
    if r=0 then AC.singleton_transition unlocked loc_source i locked loc_target 
    else if r<k then 
      let nop1=AC.single_transition unlocked loc_source i unlocked loc_target
      in let nop2S= AC.singleton_transition locked loc_source i locked loc_target
      in AC.add_transition_transitionSet nop1 nop2S
    else 
      let rel_trS= AC.singleton_transition locked loc_source i unlocked loc_target
      in let rel_tr=AC.single_transition unlocked loc_source i unlocked loc_target
      in AC.add_transition_transitionSet rel_tr rel_trS

(* Creates the set of transitions for all the threads. *)
let create_post n m k local_map =
  let create_single_tr = create_single_transition local_map m k in
  let rec create_post_aux i j r sofar =
    if i<0 then sofar
    else if j<0 then create_post_aux (i-1) (m-1) k sofar
    else if r<0 then create_post_aux i (j-1) k sofar
    else
      let trS=create_single_tr i j r in
      let next_sofar=AC.union_transitionSet_transitionSet trS sofar in
	(create_post_aux i j (r-1) next_sofar)
  in create_post_aux (n-1) (m-1) k AC.empty_transitionSet

(* preparing the arugments and starting the model checker. *)      
let start_model_checker n m k =
    let local_map=create_states m k in
    let init = AC.singleton_setOfStates unlocked (Array.make n (MapFromInts.find 0 local_map)) in
    let all_locs=MapFromInts.fold (fun j s sofar -> SetOfComponent.add s sofar) local_map SetOfComponent.empty in
    let crit = 
      let k1=k+1 in MapFromInts.fold (fun j s sofar -> if (j mod k1) !=0 
				      then SetOfComponent.add s sofar
				      else sofar) 
	              local_map SetOfComponent.empty in
    let all_shared = SetOfShared.add unlocked (SetOfShared.add locked SetOfShared.empty) in
    let error = AC.no_mutex_setOfStates crit all_locs all_shared n in
    let post = create_post n m k local_map in
    let mc_result = model_check_eager init error post in
      match mc_result with
	  Safe invariant ->print_string ("Safe. Invariant:\n"^(AC.string_of_setOfStates invariant)^".\n")
	| Unsafe trace -> print_string ("Unsafe. Error trace:\n"^(AC.string_of_trace trace)^".\n")
	
let main () =
  let error_message ()=print_string "Syntax:\nmutex <number_of_threads> <number_of_critical_sections> <length_of_critical_section>\n" in
  if ((Array.length(Sys.argv))<>4) then error_message () else
    try 
      let n=int_of_string Sys.argv.(1) and
	  m=int_of_string Sys.argv.(2) and
	  k=int_of_string Sys.argv.(3)
      in start_model_checker n m k
    with Failure f -> error_message ()
;;
main ()

